import { TargetLanguage } from "../TargetLanguage";
import { TypeGraph } from "../TypeGraph";
import { ConvenienceRenderer, ForbiddenWordsInfo } from "../ConvenienceRenderer";
import { Name, Namer } from "../Naming";
import { UnionType, Type, ClassType, EnumType } from "../Type";
import { Sourcelike } from "../Source";
import { Option } from "../RendererOptions";
export declare enum Density {
    Normal = 0,
    Dense = 1,
}
export declare enum Visibility {
    Private = 0,
    Crate = 1,
    Public = 2,
}
export default class RustTargetLanguage extends TargetLanguage {
    private readonly _denseOption;
    private readonly _visibilityOption;
    private readonly _deriveDebugOption;
    protected readonly rendererClass: new (targetLanguage: TargetLanguage, graph: TypeGraph, ...optionValues: any[]) => ConvenienceRenderer;
    constructor();
    protected getOptions(): Option<any>[];
}
export declare class RustRenderer extends ConvenienceRenderer {
    private readonly _density;
    private readonly _visibility;
    private readonly _deriveDebug;
    constructor(targetLanguage: TargetLanguage, graph: TypeGraph, leadingComments: string[] | undefined, _density: Density, _visibility: Visibility, _deriveDebug: boolean);
    protected makeNamedTypeNamer(): Namer;
    protected namerForObjectProperty(): Namer | null;
    protected makeUnionMemberNamer(): Namer | null;
    protected makeEnumCaseNamer(): Namer | null;
    protected forbiddenNamesForGlobalNamespace(): string[];
    protected forbiddenForObjectProperties(_c: ClassType, _className: Name): ForbiddenWordsInfo;
    protected forbiddenForUnionMembers(_u: UnionType, _unionName: Name): ForbiddenWordsInfo;
    protected forbiddenForEnumCases(_e: EnumType, _enumName: Name): ForbiddenWordsInfo;
    protected readonly commentLineStart: string;
    private nullableRustType;
    protected isImplicitCycleBreaker(t: Type): boolean;
    private rustType;
    private breakCycle;
    private emitRenameAttribute(propName, jsonName);
    private readonly visibility;
    protected emitStructDefinition(c: ClassType, className: Name): void;
    protected emitBlock(line: Sourcelike, f: () => void): void;
    protected emitUnion(u: UnionType, unionName: Name): void;
    protected emitEnumDefinition(e: EnumType, enumName: Name): void;
    protected emitTopLevelAlias(t: Type, name: Name): void;
    protected emitUsageExample(): void;
    protected emitSourceStructure(): void;
}
